# SPDX-FileCopyrightText: 2021-2025 Greenbone AG
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
Module for parsing and comparing Windows builds
"""

import logging
from dataclasses import dataclass

from .package import Package, PackageComparison

logger = logging.getLogger(__name__)


@dataclass
class WindowsPackage(Package):
    """Represents a Windows based package"""

    __hash__ = Package.__hash__

    def compare(self, other: "WindowsPackage") -> PackageComparison:
        if self.name != other.name:
            return PackageComparison.NOT_COMPARABLE

        return self.version_compare(self.full_version, other.full_version)

    @staticmethod
    def from_full_name(full_name: str):
        if not full_name:
            return None

        full_name = full_name.strip()
        parts = full_name.rsplit(".", 1)
        if len(parts) != 2:
            return None

        return WindowsPackage(
            full_name=full_name, name=parts[0], full_version=parts[1]
        )

    @staticmethod
    def from_name_and_full_version(name: str, full_version: str):
        if not name or not full_version:
            return None

        build = name.strip()
        revision = full_version.strip()

        full_name = f"{build}.{revision}"

        return WindowsPackage(
            full_name=full_name, name=build, full_version=revision
        )



# SPDX-FileCopyrightText: 2021-2024 Greenbone AG
#
# SPDX-License-Identifier: AGPL-3.0-or-later

import logging
import re
from abc import abstractmethod
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Callable, Dict, Optional, Set

from ...errors import PackageError

logger = logging.getLogger(__name__)


version_component_re = re.compile(r"(\d+ | .)", re.X)


class PackageType(Enum):
    RPM = "rpm"
    DEB = "deb"
    EBUILD = "ebuild"
    SLACK = "slack"
    WINDOWS = "windows"


class PackageComparison(Enum):
    EQUAL = 0  # a and b are equal
    A_NEWER = 1  # a is newer than b
    B_NEWER = 2  # b is newer than a
    # a and b are not comparable. e.g. different architectures
    NOT_COMPARABLE = 3


@dataclass
class Package:
    "Base class for different Package types"

    name: str
    full_name: str
    full_version: str

    def __guard_cmp(self, other: Any):
        if not isinstance(other, type(self)):
            raise PackageError(f"Can't compare {self!r} to {other!r}.")

    def __gt__(self, other: Any) -> bool:
        self.__guard_cmp(other)
        return self.compare(other) == PackageComparison.A_NEWER

    def __lt__(self, other: Any) -> bool:
        self.__guard_cmp(other)
        return self.compare(other) == PackageComparison.B_NEWER

    def __eq__(self, other: Any) -> bool:
        self.__guard_cmp(other)
        return self.compare(other) == PackageComparison.EQUAL

    def __ge__(self, other: Any) -> bool:
        return self.__gt__(other) or self.__eq__(other)

    def __le__(self, other: Any) -> bool:
        return self.__lt__(other) or self.__eq__(other)

    def __hash__(self) -> int:
        # allow to hash the package
        # the full name identifies the package
        return hash(self.full_name)

    @staticmethod
    def version_compare(version_a: str, version_b: str) -> PackageComparison:
        """Compares two versions"""
        if version_a == version_b:
            return PackageComparison.EQUAL

        a_parts = version_component_re.split(version_a)
        b_parts = version_component_re.split(version_b)

        for i in range(max(len(a_parts), len(b_parts))):
            if i < len(a_parts):
                a_part = a_parts[i]
            else:
                return (
                    PackageComparison.B_NEWER
                    if b_parts[i] != "~"
                    else PackageComparison.A_NEWER
                )
            if i < len(b_parts):
                b_part = b_parts[i]
            else:
                return (
                    PackageComparison.A_NEWER
                    if a_part != "~"
                    else PackageComparison.B_NEWER
                )
            if a_part == b_part:
                continue

            if a_part.isnumeric() and b_part.isnumeric():
                return (
                    PackageComparison.A_NEWER
                    if int(a_part) > int(b_part)
                    else PackageComparison.B_NEWER
                )
            if a_part.isnumeric() or b_part.isnumeric():
                return (
                    PackageComparison.A_NEWER
                    if a_part.isnumeric()
                    else PackageComparison.B_NEWER
                )

            if a_part.isalpha() and b_part.isalpha():
                return (
                    PackageComparison.A_NEWER
                    if a_part.lower() > b_part.lower()
                    else PackageComparison.B_NEWER
                )
            if a_part.isalpha():
                return (
                    PackageComparison.A_NEWER
                    if b_part == "~"
                    else PackageComparison.B_NEWER
                )
            if b_part.isalpha():
                return (
                    PackageComparison.B_NEWER
                    if a_part == "~"
                    else PackageComparison.A_NEWER
                )

            return (
                PackageComparison.A_NEWER
                if a_part != "~" and a_part > b_part or b_part == "~"
                else PackageComparison.B_NEWER
            )
        return PackageComparison.EQUAL

    @abstractmethod
    def compare(self, other: Any) -> PackageComparison:
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def from_full_name(full_name: str):
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def from_name_and_full_version(name: str, full_version: str):
        raise NotImplementedError()


@dataclass(frozen=True, unsafe_hash=True)
class PackageAdvisory:
    """Connects a package with an advisory"""

    package: Package
    oid: str
    symbol: str
    # returns None when not comparable otherwise true or false
    is_vulnerable: Callable[[Package], Optional[bool]] = field(
        compare=False, hash=False
    )


@dataclass(frozen=True)
class PackageAdvisories:
    """Container for mapping a package name to a set of advisories for this
    package"""

    package_type: PackageType
    advisories: Dict[str, Dict[str, Set[PackageAdvisory]]] = field(
        default_factory=dict
    )

    is_comparable = (  # noqa: E731
        lambda a, b: a.compare(  # pylint: disable=unnecessary-lambda-assignment
            b
        )
        != PackageComparison.NOT_COMPARABLE
    )

    comparison_map = {
        ">=": lambda a, b: (
            a > b if PackageAdvisories.is_comparable(a, b) else None
        ),
        "<=": lambda a, b: (
            a < b if PackageAdvisories.is_comparable(a, b) else None
        ),
        "=": lambda a, b: (
            a != b if PackageAdvisories.is_comparable(a, b) else None
        ),
        "<": lambda a, b: (
            a <= b if PackageAdvisories.is_comparable(a, b) else None
        ),
        ">": lambda a, b: (
            a >= b if PackageAdvisories.is_comparable(a, b) else None
        ),
    }

    def get_package_advisories_for_package(
        self, package: Package
    ) -> Dict[str, Set[PackageAdvisory]]:
        return self.advisories.get(package.name) or dict()

    def add_advisory_for_package(
        self,
        package: Package,
        advisory: str,
        verifier: Optional[str],
    ) -> None:
        if verifier not in self.comparison_map:
            verifier = ">="
        advisories = self.get_package_advisories_for_package(package)
        is_vulnerable = lambda other: self.comparison_map[  # pylint: disable=unnecessary-lambda-assignment  # noqa: E731
            verifier
        ](
            package, other
        )

        if advisory not in advisories:
            advisories[advisory] = set()

        advisories[advisory].add(
            PackageAdvisory(package, advisory, verifier, is_vulnerable)
        )
        self.advisories[package.name] = advisories

    def __len__(self) -> int:
        return len(self.advisories)



# SPDX-FileCopyrightText: 2021-2024 Greenbone AG
#
# SPDX-License-Identifier: AGPL-3.0-or-later


class NotusScannerError(Exception):
    """Base Class for error raised in Notus Scanner"""


class AdvisoriesLoadingError(NotusScannerError):
    """A problem while loading an Advisory has occurred"""


class Sha256SumLoadingError(NotusScannerError):
    """A problem while loading sha256sums has occurred"""


class PackageError(NotusScannerError):
    """Base Class for errors raised in package handling"""


class MessageParsingError(NotusScannerError):
    """A problem while parsing an incoming message"""


class ConfigFileError(NotusScannerError):
    """ "A problem while parsing the config file has occurred"""



# SPDX-FileCopyrightText: 2021-2024 Greenbone AG
#
# SPDX-License-Identifier: AGPL-3.0-or-later

import logging
from typing import Optional

from ..models.packages.package import PackageAdvisories

logger = logging.getLogger(__name__)


class AdvisoriesLoader:
    def load_package_advisories(
        self, operating_system: str
    ) -> Optional[PackageAdvisories]:
        raise NotImplementedError()



# SPDX-FileCopyrightText: 2021-2024 Greenbone AG
#
# SPDX-License-Identifier: AGPL-3.0-or-later

from datetime import datetime, timezone
from unittest import TestCase
from uuid import UUID

from notus.scanner.errors import MessageParsingError
from notus.scanner.messages.message import MessageType
from notus.scanner.messages.result import ResultMessage, ResultType


class ResultMessageTestCase(TestCase):
    def test_constructor(self):
        message = ResultMessage(
            scan_id="scan_1",
            host_ip="1.1.1.1",
            host_name="foo",
            oid="1.2.3.4.5",
            value="A Vulnerability has been found",
            uri="file://foo/bar",
        )

        self.assertIsInstance(message.message_id, UUID)
        self.assertIsInstance(message.group_id, str)
        self.assertIsInstance(message.created, datetime)

        self.assertEqual(message.message_type, MessageType.RESULT)
        self.assertEqual(message.topic, "scanner/scan/info")

        self.assertEqual(message.scan_id, "scan_1")
        self.assertEqual(message.host_ip, "1.1.1.1")
        self.assertEqual(message.host_name, "foo")
        self.assertEqual(message.oid, "1.2.3.4.5")
        self.assertEqual(message.value, "A Vulnerability has been found")

        self.assertEqual(message.result_type, ResultType.ALARM)
        self.assertEqual(message.port, "package")
        self.assertEqual(message.uri, "file://foo/bar")

    def test_serialize(self):
        created = datetime.fromtimestamp(1628512774)
        message_id = UUID("63026767-029d-417e-9148-77f4da49f49a")
        group_id = "866350e8-1492-497e-b12b-c079287d51dd"

        message = ResultMessage(
            created=created,
            message_id=message_id,
            group_id=group_id,
            scan_id="scan_1",
            host_ip="1.1.1.1",
            host_name="foo",
            oid="1.2.3.4.5",
            value="A Vulnerability has been found",
            uri="file://foo/bar",
        )

        serialized = message.serialize()
        self.assertEqual(serialized["created"], 1628512774.0)
        self.assertEqual(
            serialized["message_id"], "63026767-029d-417e-9148-77f4da49f49a"
        )
        self.assertEqual(
            serialized["group_id"], "866350e8-1492-497e-b12b-c079287d51dd"
        )
        self.assertEqual(serialized["message_type"], "result.scan")
        self.assertEqual(serialized["scan_id"], "scan_1")
        self.assertEqual(serialized["host_ip"], "1.1.1.1")
        self.assertEqual(serialized["host_name"], "foo")
        self.assertEqual(serialized["oid"], "1.2.3.4.5")
        self.assertEqual(serialized["value"], "A Vulnerability has been found")
        self.assertEqual(serialized["uri"], "file://foo/bar")
        self.assertEqual(serialized["port"], "package")
        self.assertEqual(serialized["result_type"], "ALARM")

    def test_deserialize(self):
        data = {
            "message_id": "63026767-029d-417e-9148-77f4da49f49a",
            "group_id": "866350e8-1492-497e-b12b-c079287d51dd",
            "created": 1628512774.0,
            "message_type": "result.scan",
            "scan_id": "scan_1",
            "host_ip": "1.1.1.1",
            "host_name": "foo",
            "oid": "1.2.3.4.5",
            "value": "A Vulnerability has been found",
            "uri": "file://foo/bar",
            "port": "package",
            "result_type": "ALARM",
        }

        message = ResultMessage.deserialize(data)
        self.assertEqual(
            message.message_id, UUID("63026767-029d-417e-9148-77f4da49f49a")
        )
        self.assertEqual(
            message.group_id, "866350e8-1492-497e-b12b-c079287d51dd"
        )
        self.assertEqual(
            message.created,
            datetime.fromtimestamp(1628512774.0, tz=timezone.utc),
        )
        self.assertEqual(message.message_type, MessageType.RESULT)

        self.assertEqual(message.scan_id, "scan_1")
        self.assertEqual(message.host_ip, "1.1.1.1")
        self.assertEqual(message.host_name, "foo")
        self.assertEqual(message.oid, "1.2.3.4.5")
        self.assertEqual(message.value, "A Vulnerability has been found")
        self.assertEqual(message.uri, "file://foo/bar")
        self.assertEqual(message.port, "package")
        self.assertEqual(message.result_type, ResultType.ALARM)

    def test_deserialize_invalid_message_type(self):
        data = {
            "message_id": "63026767-029d-417e-9148-77f4da49f49a",
            "group_id": "866350e8-1492-497e-b12b-c079287d51dd",
            "created": 1628512774.0,
            "message_type": "scan.status",
            "scan_id": "scan_1",
            "host_ip": "1.1.1.1",
            "host_name": "foo",
            "oid": "1.2.3.4.5",
            "value": "A Vulnerability has been found",
            "uri": "file://foo/bar",
            "port": "package",
            "result_type": "ALARM",
        }
        with self.assertRaisesRegex(
            MessageParsingError,
            "Invalid message type MessageType.SCAN_STATUS for "
            "ResultMessage. Must be MessageType.RESULT.",
        ):
            ResultMessage.deserialize(data)

    def test_deserialize_invalid_result_type(self):
        data = {
            "message_id": "63026767-029d-417e-9148-77f4da49f49a",
            "group_id": "866350e8-1492-497e-b12b-c079287d51dd",
            "created": 1628512774.0,
            "message_type": "result.scan",
            "scan_id": "scan_1",
            "host_ip": "1.1.1.1",
            "host_name": "foo",
            "oid": "1.2.3.4.5",
            "value": "A Vulnerability has been found",
            "uri": "file://foo/bar",
            "port": "package",
            "result_type": "foo",
        }

        with self.assertRaisesRegex(
            MessageParsingError,
            "error while parsing 'result_type', 'foo' is not a valid"
            " ResultType",
        ):
            ResultMessage.deserialize(data)



# SPDX-FileCopyrightText: 2021-2025 Greenbone AG
#
# SPDX-License-Identifier: AGPL-3.0-or-later

from unittest import TestCase

from notus.scanner.models.packages.windows import WindowsPackage


class WindowsPackageTestCase(TestCase):
    def test_compare_gt(self):
        package1 = WindowsPackage(
            name="10.0.22631",
            full_version="3447",
            full_name="10.0.22631.3447",
        )
        package2 = WindowsPackage(
            name="10.0.22631",
            full_version="3448",
            full_name="10.0.22631.3449",
        )

        self.assertGreater(package2, package1)

    def test_compare_gt_different_name(self):
        package1 = WindowsPackage(
            name="10.0.22631",
            full_version="3447",
            full_name="10.0.22631.3447",
        )
        package2 = WindowsPackage(
            name="10.0.22632",
            full_version="3447",
            full_name="10.0.22632.3447",
        )

        self.assertFalse(package1 > package2)
        self.assertFalse(package2 > package1)

    def test_compare_less(self):
        package1 = WindowsPackage(
            name="10.0.22631",
            full_version="3447",
            full_name="10.0.22631.3447",
        )
        package2 = WindowsPackage(
            name="10.0.22631",
            full_version="3446",
            full_name="10.0.22631.3446",
        )

        self.assertLess(package2, package1)

    def test_compare_equal(self):
        package1 = WindowsPackage(
            name="10.0.22631",
            full_version="3447",
            full_name="10.0.22631.3447",
        )
        package2 = WindowsPackage(
            name="10.0.22631",
            full_version="3447",
            full_name="10.0.22631.3447",
        )

        self.assertEqual(package2, package1)

    def test_from_full_name(self):
        package = WindowsPackage.from_full_name("")
        self.assertIsNone(package)
        package = WindowsPackage.from_full_name("3447")
        self.assertIsNone(package)
        package = WindowsPackage.from_full_name("10.0.22631.3447")
        self.assertEqual(package.name, "10.0.22631")
        self.assertEqual(package.full_version, "3447")
        self.assertEqual(package.full_name, "10.0.22631.3447")

    def test_from_name_and_full_version(self):
        package = WindowsPackage.from_name_and_full_version("", "")
        self.assertIsNone(package)
        package = WindowsPackage.from_name_and_full_version("10.0.22631", "")
        self.assertIsNone(package)
        package = WindowsPackage.from_name_and_full_version("", "3447")
        self.assertIsNone(package)
        package = WindowsPackage.from_name_and_full_version(
            "10.0.22631", "3447"
        )
        self.assertEqual(package.name, "10.0.22631")
        self.assertEqual(package.full_version, "3447")
        self.assertEqual(package.full_name, "10.0.22631.3447")
